import hashlib
import os
import subprocess
from collections import defaultdict
from pathlib import Path
from typing import Union

from neutrino_cli.compiler.templates import DockerfileTemplate, SchedulerTemplate, ConfigTemplate, MainTemplate


def hash_file(file_path: str) -> str:
    """Create a SHA256 hash of a file."""
    sha256_hash = hashlib.sha256()
    with open(file_path, 'rb') as f:
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)
    return sha256_hash.hexdigest()


def requirements_changed(build_dir: str) -> bool:
    """Check if requirements have changed since the last build."""
    requirements_path = os.path.join(build_dir, 'requirements.txt')
    hash_file_path = os.path.join(build_dir, 'requirements_hash.txt')

    # Create hash for the current requirements.txt
    current_hash = hash_file(requirements_path)

    try:
        # Read the previously stored hash
        with open(hash_file_path, 'r') as f:
            previous_hash = f.read().strip()
    except FileNotFoundError:
        previous_hash = None

    return current_hash != previous_hash


def install_requirements_from_build(build_dir: str):
    """Installs Python packages from the requirements.txt file in the build directory.

    Parameters:
        build_dir (str): Path to the build directory.
    """
    requirements_file_path = f"{build_dir}/requirements.txt"
    print(requirements_file_path)
    if os.path.exists(requirements_file_path):
        print("Installing requirements from requirements.txt...")
        subprocess.run(["pip", "install", "-r", requirements_file_path])


def merge_requirements(stock_path: str, user_path: str, output_path: str) -> None:
    """
    Merge two requirements.txt files.

    :param stock_path: File path to the stock requirements file
    :param user_path: File path to the user-defined requirements file
    :param output_path: File path to the output requirements file
    """

    def parse_requirements(file_path: str) -> defaultdict[str, Union[str, None]]:
        """
        Parse a requirements.txt file into a defaultdict.

        :param file_path: File path to the requirements file
        :return: defaultdict of package and version
        """
        reqs = defaultdict(lambda: None)
        with open(file_path, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if not line or line.startswith("#"):
                    continue
                pkg_info = line.split("==")
                reqs[pkg_info[0]] = pkg_info[1] if len(pkg_info) > 1 else None
        return reqs

    # Parse stock and user requirements
    stock_reqs = parse_requirements(stock_path)
    user_reqs = parse_requirements(user_path)

    # Update stock requirements with user requirements
    stock_reqs.update(user_reqs)

    # Write the merged requirements to the output file
    with open(output_path, 'w') as f:
        for pkg, version in stock_reqs.items():
            f.write(f"{pkg}{'==' + version if version else ''}\n")


def create_boilerplate_files(build_dir: str, boilerplate_files: list[str], root_dir: Path,
                             ignore_list: list[str] = None) -> None:
    """
    Creates boilerplate files needed for a FastAPI app.

    :param ignore_list:
    :param build_dir: Directory where the build resides
    :param boilerplate_files: List of boilerplate files to be created
    :param root_dir: Root directory of the project
    """
    for file in boilerplate_files:
        file_path = os.path.join(build_dir, file)
        with open(file_path, 'w') as f:
            if file == 'main.py':
                app_main_template = MainTemplate(root_dir, ignore_list=ignore_list)
                content = app_main_template.render()  # Assuming a render() method is available
                f.write(content)
            elif file == 'config.py':
                config_template = ConfigTemplate()
                content = config_template.render()
                f.write(content)
            elif file == 'Dockerfile':
                dockerfile_template = DockerfileTemplate()
                content = dockerfile_template.render()
                f.write(content)
            elif file == 'scheduler.py':
                scheduler_template = SchedulerTemplate()
                content = scheduler_template.render()
                f.write(content)



from neutrino_cli.compiler.templates.template import Template


class ConfigTemplate(Template):
    def __init__(self):
        template_vars = {}
        super().__init__('config_template', 'config.py.template', template_vars)

    @classmethod
    def get_template_name(cls) -> str:
        return 'config.py.template'



import os
import shutil
from pathlib import Path

from neutrino_cli.compiler.ignore_handler import should_ignore_file
from neutrino_cli.compiler.templates import InitPyTemplate
from neutrino_cli.parser.parser import compile_notebook_to_py


def create_dest_dir_if_not_exists(dest_dir: Path):
    """Create destination directory if it doesn't exist.
    Parameters:
        dest_dir (Path): Path to the destination directory.
    """
    if not dest_dir.exists():
        os.makedirs(dest_dir)


def create_init_file(dest_dir: Path, root_path: Path, build_dir: Path):
    """Create __init__.py file in a given directory.
    Parameters:
        dest_dir (Path): Path to the destination directory.
        root_path (Path): Path to the root directory.
        build_dir (Path): Path to the build directory.
    """
    if dest_dir == build_dir:
        return

    init_template = InitPyTemplate(directory=root_path)
    content = init_template.render()
    init_file_path = dest_dir / "__init__.py"
    with open(init_file_path, 'w') as f:
        f.write(content)


def copy_files(root_path: Path, dest_dir: Path, files: list[str], ignore_list: list[str]):
    """Copy or compile files to the destination directory.
    Parameters:
        root_path (Path): Path to the root directory.
        dest_dir (Path): Path to the destination directory.
        files (list): List of files to copy.
        ignore_list (list): List of files and folders to ignore.
    """
    for file in files:
        if should_ignore_file(file, ignore_list):
            continue

        src_file_path = root_path / file
        dest_file_path = dest_dir / file

        if file.endswith('.ipynb'):
            # ignore sandbox files
            if not file.endswith('sandbox.ipynb'):
                code = compile_notebook_to_py(src_file_path)
                dest_file_path = dest_file_path.with_suffix('.py')
                with open(dest_file_path, 'w') as dest_file:
                    dest_file.write(code)
        else:
            shutil.copy(src_file_path, dest_file_path)



#!/usr/bin/env python3

import os
import subprocess
import traceback
from pathlib import Path

import click
from termcolor import colored

from neutrino_cli.compiler.build_setup import hash_file, requirements_changed
from neutrino_cli.compiler.compiler import compile_notebooks_into_build, create_dest_dir_if_not_exists, format_python_files_in_dir, \
    merge_project_requirements, \
    create_boilerplate_files_in_dir
from neutrino_cli.compiler.ignore_handler import read_ignore_list
from neutrino_cli.compiler.templates import NeutrinoIgnoreTemplate


@click.group()
def cli():
    """Neutrino CLI for running and deploying projects."""
    pass


@click.command()
def init():
    """Initialize a new Neutrino project."""
    ignore_file_path = '.neutrinoignore'

    # Check if .neutrinoignore already exists
    if os.path.exists(ignore_file_path):
        print(colored(".neutrinoignore already exists.", 'yellow'))
        return

    try:
        # Create .neutrinoignore and populate with default ignores
        with open(ignore_file_path, 'w') as f:
            ignore_template = NeutrinoIgnoreTemplate()
            content = ignore_template.render()  # Assuming a render() method is available
            f.write(content)

        print(colored(f"{ignore_file_path} created successfully.", 'green'))

    except Exception as e:
        print(colored(f"An unexpected error occurred: {e}", 'red'))
        print(traceback.format_exc())


@click.command()
@click.option('--source', default=os.getcwd(), type=click.Path(exists=True), help='Path to the source folder.')
def build(source: str):
    try:
        print(colored("Building Neutrino project...", 'cyan'))

        build_dir = './build'
        create_dest_dir_if_not_exists(Path(build_dir))

        ignore_list = read_ignore_list()
        compile_notebooks_into_build(source, build_dir, ignore_list)

        format_python_files_in_dir(build_dir)
        merge_project_requirements(source, build_dir)
        create_boilerplate_files_in_dir(source, build_dir, ignore_list=ignore_list)

        # Hash the requirements.txt file
        requirements_path = os.path.join(build_dir, '../requirements.txt')
        requirements_hash = hash_file(requirements_path)

        # Save the hash for future comparison
        with open(os.path.join(build_dir, 'requirements_hash.txt'), 'w') as f:
            f.write(requirements_hash)

        print(colored("Build completed successfully.", 'green'))

    except FileNotFoundError as e:
        print(colored(f"File not found: {e}", 'red'))
    except PermissionError as e:
        print(colored(f"Permission Error: {e}", 'red'))
    except Exception as e:
        print(colored(f"An unexpected error occurred: {e}", 'red'))
        print(traceback.format_exc())
    finally:
        print(colored("Build process finished.", 'yellow'))


@click.command()
@click.option('--docker', is_flag=True, help='Run using Docker.')
@click.option('--port', default=8080, help='Port to run the server on.')
def run(docker, port):
    """Run your code."""
    build_dir = './build'
    build_dir_absolute = Path(build_dir).resolve()

    if not os.path.exists(build_dir):
        print(colored("ERROR: The project hasn't been built. Run 'build' command first.", 'red'))
        return

    main_file_path = os.path.join(build_dir, 'main.py')

    if not os.path.exists(main_file_path):
        print(colored(f"ERROR: main.py not found in {build_dir}. Ensure the project was built correctly.", 'red'))
        return

    print(colored("Launching Neutrino server...", 'cyan'))

    try:
        # Change directory to build
        os.chdir(build_dir)

        if docker:
            # Docker-related logic
            subprocess.run(["docker", "build", "-t", "neutrino-server", "."], check=True)
            subprocess.run(["docker", "run", "-p", f"{port}:8080", "neutrino-server"], check=True)
        else:
            # Check if requirements have changed
            if requirements_changed(str(build_dir_absolute)):
                print(colored("Requirements have changed, installing...", 'cyan'))
                subprocess.run(["pip", "install", "-r", "requirements.txt"], check=True)

                # Update the hash file
                with open(build_dir_absolute / 'requirements_hash.txt', 'w') as f:
                    f.write(hash_file(str(build_dir_absolute / "requirements.txt")))

            # Run FastAPI app with uvicorn for hot reloading
            subprocess.run(["uvicorn", "main:app", "--reload", "--port", str(port)], check=True)

    except subprocess.CalledProcessError as e:
        print(colored(f"Error in running the server: {e}", 'red'))
    except Exception as e:
        print(colored(f"An unexpected error occurred: {e}", 'red'))


@click.command()
def deploy():
    """Deploy your code."""
    print("Deploying code...")
    # Your 'deploy' logic here.


# Adding commands to the CLI group
cli.add_command(init)
cli.add_command(run)
cli.add_command(deploy)
cli.add_command(build)

if __name__ == "__main__":
    cli()



import re
from typing import Union

from termcolor import colored

from neutrino_cli.util.ast import get_function_name_from_ast, get_function_args_from_ast, is_async_function
from neutrino_cli.util.strings import snake_to_pascal


class HttpCell:
    def __init__(
            self,
            http: str,
            body: Union[str, list[str]],
            resp: Union[str, list[str]],
            query: Union[str, list[str]],
            headers: Union[str, list[str]],
            endpoint: str,
            func_body: str
    ):
        self.http = http
        self.endpoint = endpoint
        self.body = self._parse_fields(body)
        self.resp = self._parse_fields(resp)
        self.query = self._parse_fields(query)
        self.headers = self._parse_fields(headers)
        self.func_body = func_body

        self._check_function_args()

    def _check_function_args(self):
        func_args = get_function_args_from_ast(self.func_body)
        func_name = get_function_name_from_ast(self.func_body)
        if func_args is None:
            print(colored("WARNING: Function arguments could not be extracted from AST", 'yellow'))
            return

        if not isinstance(self.body, list) or not isinstance(self.query, list):
            print(colored(
                f"ERROR: self.body and self.query should be lists, found {type(self.body)} and {type(self.query)} instead",
                'red'))
            return

        expected_args = set([arg.split(":")[0].strip() for arg in self.body + self.query])
        url_params = set([param[0] for param in self._extract_url_params()])
        expected_args.update(url_params)
        missing_args = expected_args - set(func_args)
        if missing_args:
            print(colored(f"WARNING: Missing expected arguments in function {func_name}: {', '.join(missing_args)}",
                          'yellow'))

    @staticmethod
    def _parse_fields(fields: Union[str, list[str]]) -> list[str]:
        return fields if isinstance(fields, list) else fields.split(',') if fields else []

    def _generate_pydantic_model(self, fields: list[str], class_name: str) -> str:
        field_lines = [self._field_to_py_str(field) for field in fields]
        return f"class {class_name}(BaseModel):\n" + "\n".join(field_lines)

    @staticmethod
    def _field_to_py_str(field: str) -> str:
        if ":" not in field:
            print(colored(f"WARNING: No type hint provided for field: {field}. Defaulting to 'Any'.", 'yellow'))
            name = field.strip()
            type_ = 'Any'
        else:
            name, type_ = field.split(":")

        is_optional = '?' in type_
        type_ = type_.replace('?', '').replace('!', '').strip()
        return f"    {name.strip()}: {'Union[' + type_ + ', None]' if is_optional else type_}"

    def _extract_url_params(self) -> list[tuple[any, str]]:
        return [(name, 'str') for name in re.findall(r"/{([\w]+)}", self.endpoint)]

    def _generate_func_body_params(self) -> str:
        def clean_field(field: str) -> str:
            return field.split(":")[0].strip()

        # Handling body params
        body_params = []
        if self.body:
            for field in self.body:
                clean_field_name = clean_field(field)
                body_params.append(f"{clean_field_name}=body.{clean_field_name}")

        # Handling query params
        query_params = [f"{clean_field(field)}={clean_field(field)}" for field in self.query]

        # Handling URL params
        url_params = [f"{name}={name}" for name, _ in self._extract_url_params()]

        all_params = body_params + query_params + url_params

        return ', '.join(all_params)

    def __str__(self) -> str:
        if not self.http or not self.endpoint:
            return "# Missing HTTP method or endpoint."

        func_name = get_function_name_from_ast(self.func_body)
        pascal_func_name = snake_to_pascal(func_name) if func_name else None
        model_names = [f"{pascal_func_name}{suffix}" if pascal_func_name else suffix for suffix in
                       ["RequestBody", "ResponseModel"]]

        endpoint_def = []
        for model_name, fields in zip(model_names, [self.body, self.resp]):
            if fields:
                endpoint_def.append(f"{self._generate_pydantic_model(fields, model_name)}\n")

        endpoint_def.append(f"@router.{self.http.lower()}('{self.endpoint}')")

        url_params = self._extract_url_params()
        func_params = [f"{name}: {type_}" for name, type_ in url_params]

        if self.body:
            func_params.append(f"body: {model_names[0]}")
        if self.query:
            func_params.extend([f"{name}: {type_}" for name, type_ in [q.split(':') for q in self.query]])

        is_async = is_async_function(self.func_body)

        endpoint_def.append(f"async def {func_name}_endpoint({', '.join(func_params)}):")
        endpoint_def.extend([
            "    try:",
            f"        return {'await ' if is_async else ''}{func_name}({self._generate_func_body_params()})",
            "    except Exception as e:",
            "        if hasattr(e, 'status_code') and hasattr(e, 'message'):",
            "            raise HTTPException(status_code=e.status_code, detail=f'{e.message}')",
            "        else:",
            "            raise HTTPException(status_code=500, detail=f'Internal Server Error: {str(e)}')"
        ])
        endpoint_def.append(self.func_body)
        return "\n".join(endpoint_def)



import os
from fnmatch import fnmatchcase


def read_ignore_list(ignore_file_path: str = '.neutrinoignore') -> list[str]:
    """Read the ignore list from a given file.
    Parameters:
        ignore_file_path (str): Path to the ignore file. Defaults to '.neutrinoignore'.
    """
    if not os.path.exists(ignore_file_path):
        return []

    with open(ignore_file_path, 'r') as f:
        return [line.strip() for line in f.readlines() if line.strip()]


def should_ignore_file(file_path: str, ignore_list: list[str]) -> bool:
    """Check if a file should be ignored based on the ignore list.

    Parameters:
        file_path (str): Path to the file.
        ignore_list (list): List of patterns to ignore.

    Returns:
        bool: True if the file should be ignored, False otherwise.
    """

    # Remove comment lines from the ignore list
    filtered_ignore_list = [pattern for pattern in ignore_list if not pattern.startswith("#")]

    # Process negated patterns (those starting with "!")
    negated_patterns = [pattern[1:] for pattern in filtered_ignore_list if pattern.startswith("!")]
    filtered_ignore_list = [pattern for pattern in filtered_ignore_list if not pattern.startswith("!")]

    for pattern in negated_patterns:
        if fnmatchcase(file_path, pattern):
            return False

    for pattern in filtered_ignore_list:
        # Remove trailing slashes for directory matching
        normalized_pattern = pattern.rstrip('/')
        normalized_file_path = file_path.rstrip('/')

        if fnmatchcase(normalized_file_path, normalized_pattern) or fnmatchcase(file_path, pattern):
            return True

    return False



import re
from typing import Union

import autopep8
import nbformat
import yaml
from termcolor import colored

from neutrino_cli.parser.cells.scheduled_cell import ScheduledCell
from neutrino_cli.parser.cells.websocket_cell import WebSocketCell
from .cells import Cell, CodeCell, HttpCell



def clean_source(lines: list[str]) -> tuple[list[str], list[str]]:
    """
    Remove empty lines and split the cell into declaration lines and source lines.
    Only the first 'chunk' of comments or first multi-line comment is considered as declaration.
    """
    declaration_lines = []
    source_lines = []
    inside_multi_line = False
    first_chunk_collected = False

    for line in lines:
        stripped_line = line.strip()

        if not first_chunk_collected:
            if stripped_line.startswith('"""') or stripped_line.startswith("'''"):
                inside_multi_line = not inside_multi_line
                if not inside_multi_line:
                    first_chunk_collected = True
                continue

            if inside_multi_line or stripped_line.startswith("#"):
                declaration_lines.append(stripped_line.lstrip('#').strip())
            else:
                first_chunk_collected = True
                source_lines.append(line)  # Keep original indentation in source lines
        else:
            source_lines.append(line)  # Keep original indentation in source lines

    return declaration_lines, source_lines


def parse_cell(cell_content: dict, filepath: str) -> Union[Cell, None]:
    """
    Parses a Jupyter notebook cell to determine its type and content.

    Parameters:
    - cell_content (dict): A dictionary containing the cell content. Expects 'source' to be a key in the dict.
    - filepath (str): The path of the file containing the cell, used for error reporting.

    Returns:
    - Union[Cell, None]: Returns an object of type HttpCell, WebSocketCell, ScheduledCell, or CodeCell based on the
                   cell content. Returns None if the cell is not one of these types.
    """
    source = cell_content['source']
    lines = source.split("\n")
    cleaned_declaration_lines, source_lines = clean_source(lines)

    if not cleaned_declaration_lines and not source_lines:
        return None

    first_line = cleaned_declaration_lines[0] if cleaned_declaration_lines else None

    if first_line and re.match(r'@HTTP', first_line):
        if first_line.strip() == '@HTTP':
            cleaned_declaration_lines.pop(0)  # Remove the line entirely if it only contains @HTTP
        else:
            cleaned_declaration_lines[0] = first_line.replace('@HTTP ', '')  # Remove @HTTP but keep the rest
        return parse_http_cell(cleaned_declaration_lines, source_lines, filepath=filepath)

    elif first_line and re.match(r'@WS', first_line):
        return parse_websocket_cell(cleaned_declaration_lines, source_lines, filepath=filepath)

    elif first_line and re.match(r'@SCHEDULE', first_line):
        return parse_scheduled_cell(cleaned_declaration_lines, source_lines, filepath=filepath)

    else:
        return CodeCell(source=source)


def split_types(s: str) -> list[str]:
    """
    Split a string of types into a list of types.
    Example: "int, str, list[int]" -> ["int", "str", "list[int]"]
    :param s:
    :return:
    """
    stack = []
    start = 0
    result = []
    for i, c in enumerate(s):
        if c in ['[', '{']:
            stack.append(c)
        elif c in [']', '}']:
            stack.pop()
        elif c == ',':
            if not stack:
                result.append(s[start:i].strip())
                start = i + 1
    result.append(s[start:].strip())
    return result


def parse_http_cell(declaration_lines: list[str], source_lines: list[str], filepath: str) -> Union['HttpCell', None]:
    http_verb, endpoint = None, None
    cell_dict = {}

    first_line = declaration_lines[0]
    for x in ['GET', 'POST', 'PUT', 'DELETE', 'PATCH']:
        if x in first_line:
            http_verb = x
            endpoint = first_line.replace(x, '').strip()
            declaration_lines.pop(0)
            break

    try:
        parsed_yaml = yaml.safe_load("\n".join(declaration_lines))
        if parsed_yaml:
            cell_dict.update(parsed_yaml)
    except yaml.YAMLError as e:
        print(colored(f"Error in {filepath}:\nFailed to parse YAML: {e}", "red"))
        return None

    for key in ['body', 'resp']:
        if key in cell_dict:
            if isinstance(cell_dict[key], str):
                cell_dict[key] = split_types(cell_dict[key])

    func_body = "\n".join(source_lines).strip()

    if func_body and (http_verb and endpoint):
        try:
            return HttpCell(
                http=http_verb,
                body=cell_dict.get('body'),
                resp=cell_dict.get('resp'),
                query=cell_dict.get('query'),
                headers=cell_dict.get('headers'),
                endpoint=endpoint,
                func_body=func_body
            )
        except Exception as e:
            print(colored(f"Error parsing cell in {filepath}:\n{e}", "red"))
            return None


def parse_websocket_cell(declaration_lines: list[str], source_lines: list[str], filepath: str) -> Union[WebSocketCell, None]:
    endpoint = None
    cell_dict = {}

    first_line = declaration_lines[0]
    if "@WS" in first_line:
        endpoint = first_line.replace("@WS", '').strip()
        declaration_lines.pop(0)

    try:
        parsed_yaml = yaml.safe_load("\n".join(declaration_lines))
        if parsed_yaml:
            cell_dict.update(parsed_yaml)
    except yaml.YAMLError as e:
        print(f"Error in {filepath}:\nFailed to parse YAML: {e}")
        return None

    for key in ['query', 'headers', 'message']:
        if key in cell_dict:
            if isinstance(cell_dict[key], str):
                cell_dict[key] = split_types(cell_dict[key])

    validate_flag = cell_dict.get('validate', False)

    func_body = "\n".join(source_lines).strip()

    if func_body and endpoint:
        try:
            return WebSocketCell(
                endpoint=endpoint,
                func_body=func_body,
                ws_type=cell_dict.get('type', 'event'),
                message_schema=cell_dict.get('message'),
                query=cell_dict.get('query'),
                headers=cell_dict.get('headers'),
                validate_message_schema=validate_flag,
            )
        except Exception as e:
            print(f"Error parsing cell in {filepath}:\n{e}")
            return None


def parse_scheduled_cell(declaration_lines: list[str], source_lines: list[str], filepath: str) -> Union[ScheduledCell, None]:
    cron = None
    interval = None

    # Remove the first line as it is "@SCHEDULE"
    if declaration_lines[0].strip() == "@SCHEDULE":
        declaration_lines.pop(0)

    for line in declaration_lines:
        if ':' in line:
            key, value = map(str.strip, line.split(':', 1))
            if key == 'cron':
                cron = value
            elif key == 'interval':
                interval = value
        else:
            print(colored(f"Warning in {filepath}:\nInvalid line in schedule declaration: {line}", "red"))

    func_body = "\n".join(source_lines).strip()

    if func_body and (cron or interval):
        try:
            return ScheduledCell(
                func_body=func_body,
                cron=cron,
                interval=interval
            )
        except Exception as e:
            print(colored(f"Error parsing cell in {filepath}:\n{e}", "red"))
            return None


def parse_notebook_cells(filepath: str) -> list[str]:
    try:
        with open(filepath, 'r', encoding='utf-8') as f:
            notebook = nbformat.read(f, as_version=4)
    except FileNotFoundError:
        print(colored(f"Error: Notebook file not found: {filepath}", 'red'))
        return []
    except nbformat.reader.NotJSONError:
        print(colored(f"Error: Notebook file is not a valid JSON file: {filepath}", 'red'))
        return []

    parsed_cells = []
    for cell in notebook['cells']:
        if cell['cell_type'] == 'code':
            parsed_cell = parse_cell(cell, filepath=filepath)
            if parsed_cell:
                parsed_cells.append(parsed_cell)

    # Sort so HttpCells appear last
    sorted_cells = sorted(parsed_cells, key=lambda x: isinstance(x, HttpCell))

    return [str(cell) for cell in sorted_cells]


def compile_notebook_to_py(filepath: str) -> str:
    """Compile the notebook to Python code."""
    cells = parse_notebook_cells(filepath)

    output_lines = [
        'from fastapi import APIRouter, HTTPException, WebSocket\n',
        'from pydantic import BaseModel, ValidationError\n',
        'from scheduler import scheduler\n',
        'from typing import List, Dict, Optional, Union, Any, AsyncGenerator, Callable\n',
        'import json\n',
        'router = APIRouter()\n',
    ]
    for cell in cells:
        output_lines.append(str(cell))
        output_lines.append('\n')

    return autopep8.fix_code('\n'.join(output_lines))


if __name__ == "__main__":
    code = compile_notebook_to_py('notebook.ipynb')
    print(code)



from .dockerfile_template import DockerfileTemplate
from .init_py_template import InitPyTemplate
from .main_template import MainTemplate
from .neutrinoignore_template import NeutrinoIgnoreTemplate
from .requirements_template import RequirementsTemplate
from .scheduler_template import SchedulerTemplate
from .config_template import ConfigTemplate



import os

from jinja2 import Environment, FileSystemLoader


class Template:
    def __init__(self, template_dir, template_name, template_variables):
        this_dir = os.path.dirname(os.path.abspath(__file__))
        templates_dir = os.path.join(this_dir, template_dir)
        print(f"Templates Directory: {templates_dir}")
        self.env = Environment(loader=FileSystemLoader(templates_dir))
        self.template_name = template_name
        self.template_variables = template_variables

        if os.path.exists(os.path.join(templates_dir, template_name)):
            print(f"Template {template_name} exists.")
        else:
            print(f"Template {template_name} does not exist.")

    @classmethod
    def get_template_name(cls) -> str:
        raise NotImplementedError

    def render(self):
        template = self.env.get_template(self.template_name)
        return template.render(self.template_variables)



from pathlib import Path

from neutrino_cli.compiler.templates.template import Template
from neutrino_cli.util.strings import to_snake_case


class InitPyTemplate(Template):
    def __init__(self, directory: Path):
        nested_routers = []
        route_files = []

        # For files directly in the directory
        root_files = [
            file for file in directory.glob("*.ipynb")
            if file.stem != "__init__" and not file.name.endswith("sandbox.ipynb")
        ]
        for root_file in root_files:
            filename = to_snake_case(root_file.stem)
            router_name = f"{filename}_router"
            url_prefix = f'/{filename}' if not filename.lower().endswith('routes') else ''
            route_files.append((filename, router_name, url_prefix))

        # For subdirectories
        for subdir in directory.iterdir():
            if subdir.is_dir():
                init_file = subdir / "__init__.py"
                sub_nested_routers = [file for file in subdir.glob("*.ipynb")]

                if init_file.exists() or sub_nested_routers:
                    subdir_name = to_snake_case(subdir.name)
                    router_name = f'{subdir_name}_router'
                    url_prefix = f'/{subdir_name}' if not subdir_name.lower().endswith('routes') else ''
                    nested_routers.append((subdir_name, router_name, url_prefix))

        template_vars = {
            "nested_routers": nested_routers,
            "route_files": route_files,
        }

        super().__init__('init_py_template', 'init.py.template', template_vars)



import os
from pathlib import Path

import autopep8

from neutrino_cli.compiler.build_setup import merge_requirements, create_boilerplate_files
from neutrino_cli.compiler.file_utilities import create_init_file, create_dest_dir_if_not_exists, copy_files
from neutrino_cli.compiler.ignore_handler import should_ignore_file
from neutrino_cli.compiler.templates import RequirementsTemplate


def compile_notebooks_into_build(source_path: str, build_dir: str, ignore_list: list[str] = None):
    source_path = Path(source_path).resolve()
    build_dir = Path(build_dir).resolve()
    if ignore_list is None:
        ignore_list = []

    for root, dirs, files in os.walk(source_path):
        root_path = Path(root).resolve()

        # Make root_path relative to source_path for proper ignoring
        relative_root_path = root_path.relative_to(source_path)

        # Explicitly ignore .ipynb_checkpoints directories
        dirs[:] = [d for d in dirs if d != '.ipynb_checkpoints' and d not in ignore_list]
        # Skip directories in ignore_list
        dirs[:] = [d for d in dirs if not should_ignore_file(str((relative_root_path / d).as_posix()), ignore_list)]

        process_directory(root_path, source_path, build_dir, files, ignore_list)


def process_directory(root_path: Path, source_path: Path, build_dir: Path, files: list[str], ignore_list: list[str]):
    rel_root = root_path.relative_to(source_path)
    dest_dir = build_dir / rel_root

    if should_ignore_file(str(rel_root), ignore_list):
        return

    create_dest_dir_if_not_exists(dest_dir)

    # Skip creating __init__.py in the root directory
    if rel_root and any(file.endswith('.ipynb') for file in files):
        create_init_file(dest_dir, root_path, build_dir=build_dir)

    copy_files(root_path, dest_dir, files, ignore_list)


def format_python_files_in_dir(directory: str):
    for root, _, files in os.walk(directory):
        for file in files:
            if file.endswith('.py'):
                file_path = os.path.join(root, file)
                with open(file_path, 'r') as f:
                    code = f.read()

                formatted_code = autopep8.fix_code(code)

                with open(file_path, 'w') as f:
                    f.write(formatted_code)


def merge_project_requirements(source_path: str, build_dir: str):
    user_requirements = os.path.join(source_path, 'requirements.txt')
    output_requirements = os.path.join(build_dir, 'requirements.txt')

    # create new requirements.txt
    requirements_template = RequirementsTemplate()
    requirements_content = requirements_template.render()
    with open(output_requirements, 'w') as f:
        f.write(requirements_content)

    if os.path.exists(user_requirements):
        merge_requirements(output_requirements, user_requirements, output_requirements)


def create_boilerplate_files_in_dir(source_path: str, build_dir: str, ignore_list: list[str] = None):
    boilerplate_files = ['main.py', 'config.py', 'Dockerfile', 'scheduler.py']
    create_boilerplate_files(build_dir, boilerplate_files, root_dir=Path(source_path), ignore_list=ignore_list)



import ast
from typing import Union


def get_function_name_from_ast(code: str) -> Union[str, None]:
    """Extract the function name from the given code using AST."""
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return node.name
    return None


def get_function_param_types_from_ast(code: str) -> dict[str, str]:
    """Extract parameter types from function definition."""
    param_types = {}
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            for arg in node.args.args:
                annotation = arg.annotation
                if annotation:
                    param_types[arg.arg] = annotation.id  # assuming the annotation is simple (e.g., "int", "str")
    return param_types


def get_function_args_from_ast(code: str) -> Union[list[str], None]:
    """
    Extract the arguments of a function from the given code using AST.

    Parameters:
        code (str): The code containing the function definition.

    Returns:
        List[str]: A list of argument names if the function is found, None otherwise.
    """
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return [arg.arg for arg in node.args.args]
    return None


def is_valid_function(code: str) -> bool:
    """Check if the given code contains a valid function."""
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return True
    return False


def is_async_function(code: str) -> bool:
    """Check if the given code contains an async function."""
    tree = ast.parse(code)
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            return isinstance(node, ast.AsyncFunctionDef)
    return False



import re


def snake_to_pascal(snake_str: str) -> str:
    """
    Converts a snake_case string to PascalCase.

    Args:
        snake_str (str): The snake_case string to convert.

    Returns:
        str: The converted PascalCase string.
    """
    components = snake_str.split('_')
    return ''.join(x.title() for x in components)


def to_snake_case(string: str) -> str:
    """
    Converts a string to snake_case

    Args:
        string (str): The string to convert.

    Returns:
        str: The converted snake_case string.
    """
    string = re.sub(r'(?<!^)(?=[A-Z])', '_', string).lower()
    string = re.sub(r'[\s-]', '_', string).lower()
    return string



from pathlib import Path

from neutrino_cli.compiler.ignore_handler import should_ignore_file
from neutrino_cli.compiler.templates.template import Template
from neutrino_cli.util.strings import to_snake_case


class MainTemplate(Template):
    def __init__(self, root_dir: Path, ignore_list: list[str] = None):
        if ignore_list is None:
            ignore_list = []

        import_root_routers = []
        register_root_routers = []

        root_ipynbs = [file for file in root_dir.glob("*.ipynb") if not should_ignore_file(file.name, ignore_list)]
        for root_ipynb in root_ipynbs:
            ipynb_name = to_snake_case(root_ipynb.stem)
            router_name = f"{ipynb_name}_router"
            import_root_routers.append(f"from {ipynb_name} import router as {router_name}")
            register_root_routers.append(f"app.include_router({router_name})")

        for subdir in root_dir.iterdir():
            if subdir.is_dir() and not should_ignore_file(subdir.name, ignore_list):
                subdir_name = to_snake_case(subdir.name)

                dir_router_name = f"{subdir_name}_router"
                import_root_routers.append(f"from {subdir_name} import router as {dir_router_name}")
                url_prefix = f'/{subdir_name}' if subdir_name.endswith('_routes') else ''
                register_root_routers.append(f"app.include_router({dir_router_name}, prefix='{url_prefix}')")

        template_variables = {
            "import_root_routers": "\n".join(import_root_routers),
            "register_root_routers": "\n".join(register_root_routers),
        }

        super().__init__('main_template', 'main.py.template', template_variables)



from termcolor import colored

from neutrino_cli.util.ast import get_function_name_from_ast


class ScheduledCell:
    # Class-level counter to generate unique function names for code blocks
    _counter = 0

    def __init__(self, func_body: str, cron: str = None, interval: str = None):
        self.func_body = func_body
        self.cron = cron  # Expecting a string in format "second minute hour day month day_of_week"
        self.interval = interval

    def __str__(self) -> str:
        func_name = get_function_name_from_ast(self.func_body)
        is_already_function = func_name is not None

        if func_name is None:
            func_name = f"generated_func_{ScheduledCell._counter}"
            ScheduledCell._counter += 1

        schedule_def = []

        if self.cron:
            try:
                cron_fields = self.cron.split(' ')
                cron_dict = {
                    'second': cron_fields[0],
                    'minute': cron_fields[1],
                    'hour': cron_fields[2],
                    'day': cron_fields[3],
                    'month': cron_fields[4],
                    'day_of_week': cron_fields[5]
                }

                schedule_def.append(
                    f"@scheduler.scheduled_job('cron', id='{func_name}_cron', name='{func_name}_cron_job', **{cron_dict})")
            except IndexError:
                print(colored("WARNING: Invalid cron format. Not enough fields provided.", 'yellow'))
                return "# Invalid cron format"

        elif self.interval:
            schedule_def.append(
                f"@scheduler.scheduled_job('interval', id='{func_name}_interval', name='{func_name}_interval_job', seconds={self._parse_interval(self.interval)})")

        if not is_already_function:
            schedule_def.append(f"async def scheduled_{func_name}():")

        schedule_def.append("    " + self.func_body.replace("\n", "\n    "))

        return "\n".join(schedule_def)

    def _check_function_args(self):
        func_args = get_function_name_from_ast(self.func_body)
        if func_args is None:
            print(colored("WARNING: Function arguments could not be extracted from AST", 'yellow'))

    @staticmethod
    def _parse_interval(interval: str) -> int:
        """
        Parse the interval and return it in seconds.
        Only supports 's' for seconds, 'm' for minutes, and 'h' for hours.
        """
        unit = interval[-1]
        value = int(interval[:-1])
        if unit == 's':
            return value
        elif unit == 'm':
            return value * 60
        elif unit == 'h':
            return value * 3600
        else:
            print(colored(f"WARNING: Unsupported interval unit: {unit}. Defaulting to seconds.", 'yellow'))
            return value



from neutrino_cli.compiler.templates.template import Template


class NeutrinoIgnoreTemplate(Template):
    def __init__(self):
        template_variables = {}
        super().__init__('neutrinoignore_template', 'neutrinoignore.template', template_variables)

    @classmethod
    def get_template_name(cls):
        return 'neutrinoignore.template'



from setuptools import setup, find_packages

setup(
    name='neutrino-cli',
    version='0.1.2',
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        'click',
        'nbformat',
        'autopep8',
        'python-dotenv',
        'termcolor',
        'Jinja2',
        'PyYAML',
    ],
    entry_points={
        'console_scripts': [
            'neutrino=neutrino_cli:cli',  # Replace 'main' with the name of the entry function in your CLI
        ],
    },
)



from typing import List, Union
from termcolor import colored
from neutrino_cli.util.ast import get_function_name_from_ast, is_async_function

class WebSocketCell:
    def __init__(
            self,
            endpoint: str,
            func_body: str,
            ws_type: str = 'event',
            message_schema: Union[str, List[str]] = None,
            query: Union[str, List[str]] = None,
            headers: Union[str, List[str]] = None,
            validate_message_schema: bool = True,
    ):
        self.endpoint = endpoint
        self.ws_type = ws_type
        self.message_schema = self._parse_fields(message_schema)
        self.query = self._parse_fields(query)
        self.headers = self._parse_fields(headers)
        self.func_body = func_body
        self.validate_message_schema = validate_message_schema

    @staticmethod
    def _parse_fields(fields: Union[str, List[str], dict[str, str]]) -> Union[List[str], dict[str, str]]:
        if isinstance(fields, list):
            return fields
        elif isinstance(fields, str):
            return fields.split(',') if fields else []
        elif isinstance(fields, dict):
            return fields
        else:
            return []

    def _generate_func_body_params(self) -> str:
        params = []
        if self.query:
            for query_param in self.query:
                key, type_hint = query_param.split(':')
                params.append(f"{key}={key}")

        return ', '.join(params)

    def _generate_code(self) -> List[str]:
        """Generate code based on ws_type."""
        func_name = get_function_name_from_ast(self.func_body)
        if self.ws_type == 'event':
            return self._generate_event_code(func_name)
        elif self.ws_type == 'stream' and not self.message_schema:
            return self._generate_stream_code_no_message(func_name)
        elif self.ws_type == 'stream' and self.message_schema:
            return self._generate_stream_code_with_message(func_name)
        else:
            raise ValueError(f"Unsupported ws_type: {self.ws_type}")

    def _generate_event_code(self, func_name: str) -> List[str]:
        code = [
            "    await websocket.accept()",
            "    try:",
            "        while True:",
            f"            event_data = await websocket.receive_text()",
            f"            response = {'await ' if is_async_function(self.func_body) else ''}{func_name}(event_data, {self._generate_func_body_params()})",
            "            if response is not None:",
            "                await websocket.send_text(json.dumps(response))",
        ]
        code = code + self._handle_exception()

        return code

    def _generate_stream_code_no_message(self, func_name: str) -> List[str]:
        code = self._generate_streaming_wrapper_no_input()

        code = code + [
            "    await websocket.accept()",
            "    try:",
            f"        async for response in _streaming_wrapper(websocket, {func_name}):",
            "            await websocket.send_text(json.dumps(response))"
        ]
        code = code + self._handle_exception()

        return code

    def _generate_stream_code_with_message(self, func_name: str) -> List[str]:
        """
        Generate the code for streaming with message validation.

        Parameters:
            func_name (str): Function name to call.

        Returns:
            List[str]: List of code lines.
        """
        code = self._generate_streaming_wrapper()

        if self.validate_message_schema:
            schema_validation_code = [
                "        parsed_data = MessageSchema.parse_raw(new_input)",
                "        user_input = parsed_data.dict()",
                "    except ValidationError as e:",
                '        await websocket.send_text(f"Error: {e.json()}")',
                "        continue"
            ]
            schema_validation_code = [f"        {line}" for line in schema_validation_code]

            # Find the index where "user_input = new_input" should be
            idx = next((i for i, line in enumerate(code) if line.strip() == "user_input = new_input"), None)

            if idx is not None:
                # Replace "user_input = new_input" with validation code
                code[idx:idx + 1] = schema_validation_code

        code += [
            "    await websocket.accept()",  # Accept the WebSocket connection
            "    try:",
            f"        async for data in _streaming_wrapper(websocket, {func_name}):",
            "            await websocket.send_text(json.dumps(data))"  # Send the data received from the generator
        ]
        code += self._handle_exception()  # Add exception handling

        return code

    def _generate_pydantic_model(self, fields: list[str], class_name: str) -> str:
        field_lines = [self._field_to_py_str(field) for field in fields]
        return f"class {class_name}(BaseModel):\n" + "\n".join(field_lines)

    @staticmethod
    def _field_to_py_str(field: str) -> str:
        if ":" not in field:
            print(colored(f"WARNING: No type hint provided for field: {field}. Defaulting to 'Any'.", 'yellow'))
            name = field.strip()
            type_ = 'Any'
        else:
            name, type_ = field.split(":")

        is_optional = '?' in type_
        type_ = type_.replace('?', '').replace('!', '').strip()
        return f"    {name.strip()}: {'Union[' + type_ + ', None]' if is_optional else type_}"

    @staticmethod
    def _generate_streaming_wrapper_no_input() -> list[str]:
        code = """
async def _streaming_wrapper(websocket: WebSocket, udf: Callable[[], Any]) -> AsyncGenerator[str, None]:
    while True:
        async for data in udf():
            yield data
        """.split("\n")

        code = ["    " + line for line in code]
        return code

    def _generate_streaming_wrapper(self) -> list[str]:
        code = f"""
async def _streaming_wrapper(websocket: WebSocket, udf: Callable[[str], Any]) -> AsyncGenerator[str, None]:
    user_input = {'{}' if self.validate_message_schema else "''"}  # default value

    while True:
        try:
            new_input = await asyncio.wait_for(websocket.receive_text(), timeout=1.0)
            user_input = new_input
        except asyncio.TimeoutError:
            pass

        async for data in udf(user_input):
            yield data
        """.split("\n")

        code = ["    " + line for line in code]
        return code

    @staticmethod
    def _handle_exception() -> List[str]:
        return [
            "    except Exception as e:",
            "        if hasattr(e, 'status_code') and hasattr(e, 'message'):",
            "            await websocket.close(code=e.status_code)",
            "        else:",
            "            print(f'WebSocket Error: {e}')",
            "            await websocket.close(code=1007)"
        ]

    def __str__(self) -> str:
        func_name = get_function_name_from_ast(self.func_body)

        endpoint_def = [f"@router.websocket('{self.endpoint}')"]
        func_params = [f"websocket: WebSocket"]

        # If message_schema is present and validate_message_schema is True
        if self.message_schema and self.validate_message_schema:
            message_schema = self._generate_pydantic_model(self.message_schema, "MessageSchema")
            endpoint_def = [message_schema] + endpoint_def  # Add Pydantic schema at the beginning

        endpoint_def.append(f"async def {func_name}_websocket({', '.join(func_params)}):")
        endpoint_def.extend(self._generate_code())
        endpoint_def.append(self.func_body)

        return "\n".join(endpoint_def)
